package lcsb.mapviewer.api.drug;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.annotation.services.TaxonomyBackend;
import lcsb.mapviewer.api.QueryException;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.map.AnnotatedObject;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.persist.dao.map.ReactionDao;
import lcsb.mapviewer.persist.dao.map.species.ElementDao;
import lcsb.mapviewer.services.SecurityException;
import lcsb.mapviewer.services.interfaces.IModelService;
import lcsb.mapviewer.services.interfaces.IUserService;
import lcsb.mapviewer.services.search.ElementMatcher;
import lcsb.mapviewer.services.search.data.ElementIdentifier.ElementIdentifierType;
import lcsb.mapviewer.services.search.db.DbSearchCriteria;
import lcsb.mapviewer.services.search.db.GeneRow;
import lcsb.mapviewer.services.search.db.TargetView;
import lcsb.mapviewer.services.search.db.drug.DrugView;
import lcsb.mapviewer.services.search.db.drug.IDrugService;
import lcsb.mapviewer.services.view.AnnotationView;
import lcsb.mapviewer.services.view.AuthenticationToken;

@Transactional(value = "txManager")
public class DrugRestImpl {

	Logger								 logger	= Logger.getLogger(DrugRestImpl.class);

	@Autowired
	private IDrugService	 drugService;

	@Autowired
	private IModelService	 modelService;

	@Autowired
	private IUserService	 userService;

	@Autowired
	private ElementMatcher elementMatcher;

	public List<Map<String, Object>> getDrugsByQuery(String token, String projectId, String columns, String query) throws SecurityException, QueryException {
		AuthenticationToken authenticationToken = userService.getToken(token);
		Model model = modelService.getLastModelByProjectId(projectId, authenticationToken);
		if (model == null) {
			throw new QueryException("Project with given id doesn't exist");
		}
		Project project = model.getProject();

		Set<String> columnSet = createDrugColumnSet(columns);

		List<Map<String, Object>> result = new ArrayList<>();

		MiriamData organism = project.getOrganism();
		if (organism == null) {
			organism = TaxonomyBackend.HUMAN_TAXONOMY;
		}
		DrugView drug = drugService.getByName(query, new DbSearchCriteria().model(model).organisms(organism).colorSet(0));
		if (drug != null) {
			List<Model> models = getModels(model);
			result.add(prepareDrug(drug, columnSet, models));
		}

		return result;
	}

	private List<Model> getModels(Model model) {
		List<Model> models = new ArrayList<>();
		models.add(model);
		models.addAll(model.getSubmodels());
		return models;
	}

	/**
	 * @return the userService
	 * @see #userService
	 */
	public IUserService getUserService() {
		return userService;
	}

	/**
	 * @param userService
	 *          the userService to set
	 * @see #userService
	 */
	public void setUserService(IUserService userService) {
		this.userService = userService;
	}

	/**
	 * @return the modelService
	 * @see #modelService
	 */
	public IModelService getModelService() {
		return modelService;
	}

	/**
	 * @param modelService
	 *          the modelService to set
	 * @see #modelService
	 */
	public void setModelService(IModelService modelService) {
		this.modelService = modelService;
	}

	private Map<String, Object> prepareDrug(DrugView drug, Set<String> columnsSet, List<Model> models) {
		Map<String, Object> result = new HashMap<>();
		for (String string : columnsSet) {
			String column = string.toLowerCase();
			Object value = null;
			if (column.equals("id") || column.equals("idobject")) {
				value = drug.getName();
			} else if (column.equals("name")) {
				value = drug.getName();
			} else if (column.equals("references")) {
				value = drug.getDrugLinks();
			} else if (column.equals("description")) {
				value = drug.getDescription();
			} else if (column.equals("bloodbrainbarrier")) {
				value = drug.getBloodBrainBarrier();
			} else if (column.equals("brandnames")) {
				value = drug.getBrandNames();
			} else if (column.equals("synonyms")) {
				value = drug.getSynonyms();
			} else if (column.equals("targets")) {
				value = prepareTargets(drug.getTargetRows(), models);
			} else {
				value = "Unknown column";
			}
			result.put(string, value);
		}
		return result;
	}

	private List<Map<String, Object>> prepareTargets(Collection<TargetView> targets, List<Model> models) {
		List<Map<String, Object>> result = new ArrayList<>();
		for (TargetView target : targets) {
			result.add(prepareTarget(target, models));
		}
		return result;
	}

	private Map<String, Object> prepareTarget(TargetView target, List<Model> models) {
		Map<String, Object> result = new HashMap<>();
		result.put("name", target.getName());
		result.put("references", target.getReferences());
		List<AnnotationView> participants = new ArrayList<>();
		for (GeneRow row : target.getProteins()) {
			participants.add(row.getAnnotation());
		}
		result.put("targetParticipants", participants);

		List<Map<String, Object>> targetedObjects = new ArrayList<>();
		for (Model model : models) {
			for (AnnotatedObject object : model.getAnnotatedObjects()) {
				if (elementMatcher.elementMatch(target, object)) {
					Map<String, Object> elementMapping = new HashMap<>();

					elementMapping.put("id", object.getId());
					elementMapping.put("type", getType(object));
					elementMapping.put("modelId", model.getId());
					targetedObjects.add(elementMapping);
				}
			}
		}

		result.put("targetElements", targetedObjects);

		return result;
	}

	private String getType(AnnotatedObject object) {
		if (object instanceof Reaction) {
			return ElementIdentifierType.REACTION.getJsName();
		} else if (object instanceof Element) {
			return ElementIdentifierType.ALIAS.getJsName();
		} else {
			throw new InvalidArgumentException("Unknown type of element " + object.getClass());
		}
	}

	private Set<String> createDrugColumnSet(String columns) {
		Set<String> columnsSet = new HashSet<>();
		if (columns.equals("")) {
			columnsSet.add("name");
			columnsSet.add("references");
			columnsSet.add("description");
			columnsSet.add("bloodBrainBarrier");
			columnsSet.add("brandNames");
			columnsSet.add("synonyms");
			columnsSet.add("id");
			columnsSet.add("targets");
		} else {
			for (String str : columns.split(",")) {
				columnsSet.add(str);
			}
		}
		return columnsSet;
	}

	/**
	 * @return the drugService
	 * @see #drugService
	 */
	public IDrugService getDrugService() {
		return drugService;
	}

	/**
	 * @param drugService
	 *          the drugService to set
	 * @see #drugService
	 */
	public void setDrugService(IDrugService drugService) {
		this.drugService = drugService;
	}

	public List<Map<String, Object>> getDrugsByTarget(String token, String projectId, String targetType, String targetId, String columns)
			throws SecurityException, QueryException {
		AuthenticationToken authenticationToken = userService.getToken(token);
		Model model = modelService.getLastModelByProjectId(projectId, authenticationToken);
		if (model == null) {
			throw new QueryException("Project with given id doesn't exist");
		}
		Project project = model.getProject();

		Integer dbId = Integer.valueOf(targetId);
		List<Element> targets = new ArrayList<>();
		if (targetType.equals(ElementIdentifierType.ALIAS.getJsName())) {
			Element element = model.getElementByDbId(dbId);
			for (Model m : model.getSubmodels()) {
				if (element == null) {
					element = m.getElementByDbId(dbId);
				}
			}
			if (element == null) {
				throw new QueryException("Invalid element identifier for given project");
			}
			targets.add(element);
		} else {
			throw new QueryException("Targeting for the type not implemented");
		}
		MiriamData organism = project.getOrganism();
		if (organism == null) {
			organism = TaxonomyBackend.HUMAN_TAXONOMY;
		}

		Set<String> columnSet = createDrugColumnSet(columns);

		List<DrugView> drugs = drugService.getForTargets(targets, new DbSearchCriteria().model(model).organisms(organism));

		List<Map<String, Object>> result = new ArrayList<>();

		List<Model> models = getModels(model);

		for (DrugView drug : drugs) {
			result.add(prepareDrug(drug, columnSet, models));
		}

		return result;
	}

}
